{% extends "base.html" %}

{% block extrahead %}
<meta property="og:title" content="DeepSaki - Deep Learning Extension for TensorFlow" />
<meta property="og:description" content="DeepSaki is an add-on to TensorFlow. It provides a variaty of custom classes ranging from activation functions to entire models, helper functions to facilitate connectiong to your, compute HW and many more!" />
<meta property="og:type" content="website" />
<meta property="og:image" content="https://sascha-kirch.github.io/DeepSaki/latest/assets/images/social_card.png" />
<meta property="og:url" content="https://sascha-kirch.github.io/DeepSaki/latest/" />
<meta name="twitter:card" content="summary_large_image">
{% endblock %}
